import json

import pytest


# import vcrpy
# https://vcrpy.readthedocs.io/en/latest/advanced.html#filter-sensitive-data-from-the-request
# https://vcrpy.readthedocs.io/en/latest/advanced.html#automatic-cassette-naming
